// This file is part of the Acts project.
//
// Copyright (C) 2017-2019 CERN for the benefit of the Acts project
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

#pragma once

#include <limits>

#include "Acts/EventData/NeutralParameters.hpp"
#include "Acts/Geometry/GeometryContext.hpp"
#include "Acts/MagneticField/MagneticFieldContext.hpp"
#include "Acts/Propagator/detail/DebugOutputActor.hpp"
#include "Acts/Tests/CommonHelpers/FloatComparisons.hpp"
#include "Acts/Utilities/Helpers.hpp"

namespace tt = boost::test_tools;

namespace Acts {
namespace IntegrationTest {

using Covariance = BoundSymMatrix;

// Create a test context
GeometryContext tgContext = GeometryContext();
MagneticFieldContext mfContext = MagneticFieldContext();

/// Helper method to create a transform for a plane
/// to mimic detector situations, the plane is roughly
/// perpendicular to the track
///
/// @param nnomal The nominal normal direction
/// @param angleT Rotation around the norminal normal
/// @param angleU Roation around the original U axis
std::shared_ptr<Transform3D> createPlanarTransform(const Vector3D& nposition,
                                                   const Vector3D& nnormal,
                                                   double angleT,
                                                   double angleU) {
  // the rotation of the destination surface
  Vector3D T = nnormal.normalized();
  Vector3D U = std::abs(T.dot(Vector3D::UnitZ())) < 0.99
                   ? Vector3D::UnitZ().cross(T).normalized()
                   : Vector3D::UnitX().cross(T).normalized();
  Vector3D V = T.cross(U);
  // that's the plane curvilinear Rotation
  RotationMatrix3D curvilinearRotation;
  curvilinearRotation.col(0) = U;
  curvilinearRotation.col(1) = V;
  curvilinearRotation.col(2) = T;
  // curvilinear surfaces are boundless
  Transform3D ctransform{curvilinearRotation};
  ctransform.pretranslate(nposition);
  ctransform.prerotate(AngleAxis3D(angleT, T));
  ctransform.prerotate(AngleAxis3D(angleU, U));
  //
  return std::make_shared<Transform3D>(ctransform);
}

/// Helper method to create a transform for a plane
/// to mimic detector situations, the plane is roughly
/// perpendicular to the track
///
/// @param nnomal The nominal normal direction
/// @param angleT Rotation around the norminal normal
/// @param angleU Roation around the original U axis
std::shared_ptr<Transform3D> createCylindricTransform(const Vector3D& nposition,
                                                      double angleX,
                                                      double angleY) {
  Transform3D ctransform;
  ctransform.setIdentity();
  ctransform.pretranslate(nposition);
  ctransform.prerotate(AngleAxis3D(angleX, Vector3D::UnitX()));
  ctransform.prerotate(AngleAxis3D(angleY, Vector3D::UnitY()));
  return std::make_shared<Transform3D>(ctransform);
}

template <typename end_parameters_t, typename Propagator_type,
          typename start_parameters_t>
Vector3D constant_field_propagation(const Propagator_type& propagator,
                                    start_parameters_t start, double pT,
                                    double phi, double theta, double Bz,
                                    double disttol = 0.1 *
                                                     Acts::UnitConstants::um,
                                    bool debug = false) {
  using namespace Acts::UnitLiterals;
  namespace VH = VectorHelpers;

  // setup propagation options
  PropagatorOptions<> options(tgContext, mfContext);
  options.pathLimit = 5_m;
  options.maxStepSize = 1_cm;
  options.debug = debug;

  double q = start.charge();
  Vector3D pos = start.position();
  Vector3D mom = start.momentum();

  // do propagation
  const auto& tp = propagator.propagate(start, options).value().endParameters;

  // test propagation invariants
  // clang-format off
    CHECK_CLOSE_ABS(pT, VH::perp(tp->momentum()), 1_keV);
    CHECK_CLOSE_ABS(mom.z(), tp->momentum()(2), 1_keV);
    CHECK_CLOSE_ABS(theta, VH::theta(tp->momentum()), 1e-4);
  // clang-format on

  double r = (q * Bz != 0.) ? std::abs(pT / (q * Bz))
                            : std::numeric_limits<double>::max();

  // calculate number of turns of helix
  double turns = options.pathLimit / (2 * M_PI * r) * sin(theta);
  // respect direction of curl
  turns = (q * Bz < 0) ? turns : -turns;

  // calculate expected final momentum direction in phi in [-pi,pi]
  double exp_phi = std::fmod(phi + turns * 2 * M_PI, 2 * M_PI);
  if (exp_phi < -M_PI) {
    exp_phi += 2 * M_PI;
  }
  if (exp_phi > M_PI) {
    exp_phi -= 2 * M_PI;
  }

  // calculate expected position
  double exp_z = pos.z() + mom.z() / pT * 2 * M_PI * r * std::abs(turns);

  // calculate center of bending circle in transverse plane
  double xc, yc;
  // offset with respect to starting point
  double dx = r * cos(M_PI / 2 - phi);
  double dy = r * sin(M_PI / 2 - phi);
  if (q * Bz < 0) {
    xc = pos.x() - dx;
    yc = pos.y() + dy;
  } else {
    xc = pos.x() + dx;
    yc = pos.y() - dy;
  }
  // phi position of starting point in bending circle
  double phi0 = std::atan2(pos.y() - yc, pos.x() - xc);

  // calculated expected position in transverse plane
  double exp_x = xc + r * cos(phi0 + turns * 2 * M_PI);
  double exp_y = yc + r * sin(phi0 + turns * 2 * M_PI);

  // clang-format off
    CHECK_CLOSE_ABS(exp_phi, VH::phi(tp->momentum()), 1e-4);
    CHECK_CLOSE_ABS(exp_x, tp->position()(0), disttol);
    CHECK_CLOSE_ABS(exp_y, tp->position()(1), disttol);
    CHECK_CLOSE_ABS(exp_z, tp->position()(2), disttol);
  // clang-format on
  return tp->position();
}

template <typename forward_parameters_t, typename backward_parameters_t,
          typename Propagator_type, typename start_parameters_t>
void foward_backward(const Propagator_type& propagator, double plimit,
                     start_parameters_t start,
                     double disttol = 1 * Acts::UnitConstants::um,
                     double momtol = 10 * Acts::UnitConstants::keV,
                     bool debug = false) {
  using namespace Acts::UnitLiterals;

  // setup propagation options
  // Action list and abort list
  using DebugOutput = Acts::detail::DebugOutputActor;
  using ActionList = Acts::ActionList<DebugOutput>;

  PropagatorOptions<ActionList> fwdOptions(tgContext, mfContext);
  fwdOptions.pathLimit = plimit;
  fwdOptions.maxStepSize = 1_cm;
  fwdOptions.debug = debug;

  PropagatorOptions<ActionList> bwdOptions(tgContext, mfContext);
  bwdOptions.direction = backward;
  bwdOptions.pathLimit = -plimit;
  bwdOptions.maxStepSize = 1_cm;
  bwdOptions.debug = debug;

  const Vector3D posStart = start.position();
  const Vector3D momStart = start.momentum();

  // do forward propagation
  const auto& fwdResult =
      propagator.template propagate<forward_parameters_t>(start, fwdOptions)
          .value();

  // do the backward propagation
  const auto& bwdResult = propagator
                              .template propagate<backward_parameters_t>(
                                  *fwdResult.endParameters, bwdOptions)
                              .value();
  const Vector3D& bwdPosition = bwdResult.endParameters->position();
  const Vector3D& bwdMomentum = bwdResult.endParameters->momentum();

  // test propagation invariants
  // clang-format off
    CHECK_CLOSE_ABS(posStart.x(), bwdPosition(0), disttol);
    CHECK_CLOSE_ABS(posStart.y(), bwdPosition(1), disttol);
    CHECK_CLOSE_ABS(posStart.z(), bwdPosition(2), disttol);
    CHECK_CLOSE_ABS(momStart.x(), bwdMomentum(0), momtol);
    CHECK_CLOSE_ABS(momStart.y(), bwdMomentum(1), momtol);
    CHECK_CLOSE_ABS(momStart.z(), bwdMomentum(2), momtol);
  // clang-format on

  if (debug) {
    auto fwdOutput = fwdResult.template get<DebugOutput::result_type>();
    std::cout << ">>>>> Output for forward propagation " << std::endl;
    std::cout << fwdOutput.debugString << std::endl;
    std::cout << " - resulted at position : "
              << fwdResult.endParameters->position() << std::endl;

    auto bwdOutput = bwdResult.template get<DebugOutput::result_type>();
    std::cout << ">>>>> Output for backward propagation " << std::endl;
    std::cout << bwdOutput.debugString << std::endl;
    std::cout << " - resulted at position : "
              << bwdResult.endParameters->position() << std::endl;
  }
}

// test propagation to cylinder
template <typename Propagator_type, typename start_parameters_t>
std::pair<Vector3D, double> to_cylinder(const Propagator_type& propagator,
                                        start_parameters_t start, double plimit,
                                        double rand1, double rand2,
                                        double /*rand3*/, bool debug = false) {
  using namespace Acts::UnitLiterals;

  // setup propagation options
  PropagatorOptions<> options(tgContext, mfContext);
  // setup propagation options
  options.maxStepSize = plimit * 0.1;
  options.pathLimit = plimit;
  options.debug = debug;

  // The transform at the destination
  auto seTransform = createCylindricTransform(Vector3D(0., 0., 0.),
                                              0.04 * rand1, 0.04 * rand2);
  auto endSurface = Surface::makeShared<CylinderSurface>(
      seTransform, plimit, std::numeric_limits<double>::max());

  // Increase the path limit - to be safe hitting the surface
  options.pathLimit *= 2;

  const auto result = propagator.propagate(start, *endSurface, options).value();
  const auto& tp = result.endParameters;
  // check for null pointer
  BOOST_CHECK(tp != nullptr);
  // The position and path length
  return std::pair<Vector3D, double>(tp->position(), result.pathLength);
}

// test propagation to most surfaces
template <typename Propagator_type, typename Surface_type,
          typename start_parameters_t>
std::pair<Vector3D, double> to_surface(const Propagator_type& propagator,
                                       start_parameters_t start, double plimit,
                                       double rand1, double rand2, double rand3,
                                       bool planar = true, bool debug = false) {
  using namespace Acts::UnitLiterals;
  using DebugOutput = detail::DebugOutputActor;

  // setup propagation options
  PropagatorOptions<ActionList<DebugOutput>> options(tgContext, mfContext);
  // setup propagation options
  options.maxStepSize = plimit;
  options.pathLimit = plimit;
  options.debug = debug;

  const auto result_s = propagator.propagate(start, options).value();
  const auto& tp_s = result_s.endParameters;
  // The transform at the destination
  auto seTransform = planar
                         ? createPlanarTransform(tp_s->position(),
                                                 tp_s->momentum().normalized(),
                                                 0.1 * rand3, 0.1 * rand1)
                         : createCylindricTransform(tp_s->position(),
                                                    0.04 * rand1, 0.04 * rand2);

  auto endSurface = Surface::makeShared<Surface_type>(seTransform, nullptr);
  // Increase the path limit - to be safe hitting the surface
  options.pathLimit *= 2;

  if (debug) {
    std::cout << ">>> Path limit for this propgation is set to: "
              << options.pathLimit << std::endl;
  }

  auto result = propagator.propagate(start, *endSurface, options);
  const auto& propRes = *result;
  const auto& tp = propRes.endParameters;
  // check the result for nullptr
  BOOST_CHECK(tp != nullptr);

  // screen output in case you are running in debug mode
  if (debug) {
    const auto& debugOutput = propRes.template get<DebugOutput::result_type>();
    std::cout << ">>> Debug output of this propagation " << std::endl;
    std::cout << debugOutput.debugString << std::endl;
    std::cout << ">>> Propagation status is : ";
    if (result.ok()) {
      std::cout << "success";
    } else {
      std::cout << result.error();
    }
    std::cout << std::endl;
  }

  // The position and path length
  return std::pair<Vector3D, double>(tp->position(), propRes.pathLength);
}

template <typename end_parameters_t, typename Propagator_type, typename start_parameters_t>
auto covariance_curvilinear(const Propagator_type& propagator, start_parameters_t start,
                                  double plimit, bool debug = false) {
  using namespace Acts::UnitLiterals;

  // setup propagation options
  DenseStepperPropagatorOptions<> options(tgContext, mfContext);
  options.maxStepSize = plimit;
  options.pathLimit = 0.1 * plimit;
  options.debug = debug;
  options.tolerance = 1e-9;

  const auto result = propagator.template propagate<end_parameters_t>(start, options).value();
  const auto& tp = result.endParameters;

  return *(tp->covariance());
}

template <typename Propagator_type, typename DestSurface_type, typename start_parameters_t>
Covariance covariance_bound(const Propagator_type& propagator,
                            double plimit, double rand1, double rand2,
                            double rand3, start_parameters_t start,
                            bool destPlanar = true, bool debug = false) {
  using namespace Acts::UnitLiterals;

  // setup propagation options
  DenseStepperPropagatorOptions<> options(tgContext, mfContext);
  options.maxStepSize = 0.1 * plimit;
  options.pathLimit = plimit;
  options.debug = debug;


  // create curvilinear start parameters
  CurvilinearParameters start_c(std::nullopt, start.position(), start.momentum(), start.charge(), start.time());
  const auto result_c = propagator.propagate(start_c, options).value();
  const auto& tp_c = result_c.endParameters;

  // create destination surface
  auto seTransform = destPlanar
                         ? createPlanarTransform(tp_c->position(),
                                                 tp_c->momentum().normalized(),
                                                 0.05 * rand3, 0.05 * rand1)
                         : createCylindricTransform(tp_c->position(),
                                                    0.01 * rand1, 0.01 * rand2);
  auto endSurface = Surface::makeShared<DestSurface_type>(seTransform, nullptr);
  
  // increase the path limit - to be safe hitting the surface
  options.pathLimit *= 2;

  const auto result = propagator.propagate(start, *endSurface, options).value();
  const auto& tp = result.endParameters;

  // get obtained covariance matrix
  return *(tp->covariance());
}
}  // namespace IntegrationTest
}  // namespace Acts
